#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include"spline.h"

void spline3make(int n, double *f, double *c)
{
  double *alpha=malloc(n*sizeof(double));
  double *beta=malloc(n*sizeof(double));
  int i;

  alpha[1]=-2.;
   beta[1]=(f[2]+4*f[1]-5*f[0])/2.;
  for(i=1;i<n-1;i++)
  { alpha[i+1]=-1./(4.+alpha[i]);
     beta[i+1]=alpha[i+1]*(beta[i]-3*(f[i+1]-f[i-1]));
  }

  c[n-1]=(5*f[n-1]-4*f[n-2]-f[n-3]-4*beta[n-1])/(4*alpha[n-1]+2.);

  for(i=n-2;i>=0;i--) c[i]=alpha[i+1]*c[i+1]+beta[i+1];

  free(alpha); free(beta);
}

double spline3val(int n, double * f, double * c, double x, double * deriv)
{
   int  n0=(int)x;
   double dx=x-n0;
   double dx1=1.-dx;

if(deriv) *deriv=
   6*dx*dx1*(f[n0+1]-f[n0])+dx1*(1.-3*dx)*c[n0]+ dx*(1-3*dx1)*c[n0+1];

return f[n0]*dx1*dx1*(1.+2*dx)
     + f[n0+1]*dx*dx*(1.+2*dx1)
     + dx*dx1*( dx1*c[n0] - dx*c[n0+1]);
}

static void c_and_d(int I,double *f,double *b,double *c,double *d)
{  int i;

   c[0]= 6*(f[1]-f[0])-4*b[0]-2*b[1];

   for(i=1;i<I;i++) 
   { 
      c[i]=-6*(f[i]-f[i-1])+2*b[i-1]+4*b[i];
      d[i]=c[i]-c[i-1]; 
   }
   d[0]=d[1];
}




/* +++++++++++++++++++++++++++ INTEGRALS +++++++++++++++++++++++++++++++ */
static void INTEGRALS(int N,double p, double *Iexp,double *dIexp,double *I,
	       double **A,double **F3,double *b,double *c,double *d)
{ int i,j;
  double *f;
  
  f=(double *)malloc((N+2) *sizeof(double));


  for(j=0;j<=N+1;j++)
       f[j]=0;

/*------------- Iexp[i]/dIexp[i] -------------------*/
 for(i=1;i<=N+1;i++)
 { if(dIexp[i-1]==0)  dIexp[i-1]=1.e-09;
   I[i]=Iexp[i-1]/dIexp[i-1];
 }
/*--------------------------------------------------*/

for(i=0;i<=N+1;i++)
{
  if(i==0)
   { f[0]=1;
     f[1]=1./8.;
    }
  if(i==1)
   { f[0]=0;
     f[1]=1;
     f[2]=8./37.;
    }
  if(i==2)
   { f[2]=1;
     f[1]=1./8.;
     f[3]=1./4.;
    }
  if(i==3)
   { f[3]=1;
     f[2]=9./37.;
     f[4]=1./4.;
    }

  if( (i>=4)&&(i<=(N-3)) )
   { f[i]=1;
     f[i-1]=1./4.;
     f[i+1]=1./4.;
    }

  if(i==(N-2))
   { f[N-2]=1;
     f[N-1]=9./37.;
     f[N-3]=1./4.;
    }
  if(i==(N-1))
   { f[N-1]=1;
     f[N]=1./8.;
     f[N-2]=1./4.;
    }
  if(i==N)
   { f[N]=1;
     f[N-1]=8./37.;
    }
  if(i==(N+1))
   { f[N+1]=1;
     f[N]=1./8.;
    }

    spline3make(N+2,f,b);
    c_and_d(N+2,f,b,c,d);

/*-----------------integrals ------------------------------------*/

 for(j=1;j<=N+1;j++)
   {
    A[i][j]=( f[j]-b[j]/2.+c[j]/6.-d[j]/24.)/dIexp[j-1];

    if(Iexp[j-1]==0) Iexp[j-1]=1.e-09;

    F3[i][j]=(d[j]-d[j-1])/pow(fabs(Iexp[j-1]),p);
   }

 if(i!=0)  f[i-1]=0;

}

 free((void*)f);

}



/* +++++++++++++++++++++ GAUSS +++++++++++++++++++++++ */
static void GAUSS(double psi,int N,double **A,double **F3,double *I,double *B,
							    double **C)
{ int g,v,i,j,k,l,s,m,n;

/*------------------C[i][j] and B[j] -----------------*/
 for(g=0;g<=N+1;g++)
  { if(g>=4)  s=g-4;
    else      s=0;

    if(g<=(N-3)) l=g+4;
    else         l=N+1;

    for(v=s;v<=l;v++)
     { C[g][v]=0;

       if(g>=2) m=g-1;
       else     m=1;

       if(g<=(N-3)) n=g+3;
       else         n=N+1;

       for(j=m;j<=n;j++)
	 C[g][v]=C[g][v]+A[g][j]*A[v][j]+psi*F3[g][j]*F3[v][j];
      }
   }


 for(g=0;g<=N+1;g++)
  { B[g]=0;

    if(g>=2) m=g-1;
    else     m=1;

    if(g<=(N-3)) n=g+2;
    else         n=N+1;

    for(j=m;j<=n;j++)
      B[g]=B[g]+A[g][j]*I[j];
   }

/*----------------- GAUSS -------------------*/
for(k=4;k>=1;k--)
 for(j=k;j<=N+1;j++)
  if(C[j-1][j-k]!=0)
  { B[j]=B[j]-B[j-1]*C[j][j-k]/C[j-1][j-k];
    if((N+1-j)>=4) s=j+4;
    else           s=N+1;
    for(i=s;i>=j-k;i--)
      C[j][i]=C[j][i]-C[j-1][i]*C[j][j-k]/C[j-1][j-k];
  }

for(k=4;k>=1;k--)
 for(j=N+1-k;j>=0;j--)
  if(C[j+1][j+k]!=0)
  { B[j]=B[j]-B[j+1]*C[j][j+k]/C[j+1][j+k];
    for(i=j;i<=j+k;i++)
      C[j][i]=C[j][i]-C[j+1][i]*C[j][j+k]/C[j+1][j+k];
  }

  for(j=0;j<=N+1;j++) B[j]=B[j]/C[j][j];
}





/* +++++++++++++++++++++ KSI_2 ++++++++++++++++++++++++++++++++++++++++*/
static double KSI_2(double psi,int N,double *Iexp,double *dIexp,double *I,double **A,
					     double **F3,double *B,double **C)
{ int i,j,m,n;
  double ksi2,*Iteor;

  Iteor=(double *)malloc((N+1) *sizeof(double));
printf("psi= %E ",psi);
/*------------ INTEGRALS end GAUSS -------------------*/
 GAUSS(psi,N,A,F3,I,B,C);


 for(i=0;i<=N;i++)  Iteor[i]=0;
 ksi2=0;
/*---------------- ksi2----------------------------------------------*/
for(i=0;i<=N;i++)
 { if(i>=2) m=i-2;
   else     m=0;

   if(i<=(N-3)) n=i+3;
   else         n=N+1;

   for(j=m;j<=n;j++)
    Iteor[i]=Iteor[i]+B[j]*A[j][i+1]*dIexp[i];
  }
for(i=0;i<=N;i++)
 ksi2=ksi2+(Iteor[i]-Iexp[i])*(Iteor[i]-Iexp[i])/(dIexp[i]*dIexp[i]);
 /*--------------------------------------------------------------------*/
printf(" chi2= %f\n",ksi2);
 free((void*)Iteor);
 return(ksi2);

}


/* ++++++++++++++++++++++ F3_2 ++++++++++++++++++++++++++++++++++ */
static double F3_2(int N,double **F3,double *B)
{ int i,j,m,n;
  double delta_f3_2,*F3teor;

  F3teor=(double *)malloc((N+1) *sizeof(double));

  for(i=0;i<=N;i++)  F3teor[i]=0;
  delta_f3_2=0;

/*---------------- f3_2--------------------*/
for(i=0;i<=N;i++)
  { if(i>=2) m=i-2;
   else     m=0;

   if(i<=(N-3)) n=i+3;
   else         n=N+1;

   for(j=m;j<=n;j++)
     F3teor[i]=F3teor[i]+B[j]*F3[j][i+1];
  }
for(i=0;i<=N;i++)
 delta_f3_2=delta_f3_2+F3teor[i]*F3teor[i];
/*-----------------------------------------*/
free((void*)F3teor);
 return(delta_f3_2);

}



/* +++++++++++++++ teor_psi +++++++++++++++++++++++++++++++++++++++++++++*/
static double teor_psi(double k,int N,
		double *Iexp,double *dIexp,double *I,double **A,double **F3,
		double *B,double **C)
{
  double k0,k1,
	 psi,psi0,psi1,
	 ksi0,ksi1,
	 ksi0_4,ksi1_4;

   psi0=0;
   psi1=1.e+05;

   ksi0=KSI_2(psi0,N,Iexp,dIexp,I,A,F3,B,C);
   ksi1=KSI_2(psi1,N,Iexp,dIexp,I,A,F3,B,C);

     ksi0_4=ksi0*ksi0*ksi0*ksi0;
     ksi1_4=ksi1*ksi1*ksi1*ksi1;

     k1=(psi1-psi0)/(ksi1_4-ksi0_4);
     k0=psi0-k1*ksi0_4;

     psi=k0+k1*(N+1)*(N+1)*(N+1)*(N+1);


  return(psi);
}


/* +++++++++++++++++++++++ SPLINE ++++++++++++++++++++++++++++++++++++++ */
void spline3smooth(double k, double p,  int N, double *Iexp,double *dIexp, double *I)
{
 int i,j;
 double ksi2,ksi_left,ksi_right,ksi_psi,delta_ksi,
	psi,psi_right,psi_left,delta_psi,
	delta,
	delta_f3_2;
 double **A;
 double **F3;
 double *B;
 double **C;

double *  c=(double *)malloc((N+1) *sizeof(double));
double *  b=(double *)malloc((N+1) *sizeof(double));
double *  d=(double *)malloc((N+1) *sizeof(double));

 A=(double **)malloc((N+1) *sizeof(double *));
 for(j=0;j<=N;j++)
  A[j]=(double *)malloc((N+1) *sizeof(double));

 F3=(double **)malloc((N+1) *sizeof(double *));
 for(j=0;j<=N;j++) F3[j]=(double *)malloc((N+1) *sizeof(double));

 B=(double *)malloc((N+1) *sizeof(double));

 C=(double **)malloc((N+1) *sizeof(double *));
 for(j=0;j<=N;j++) C[j]=(double *)malloc((N+1) *sizeof(double));


/*------------delta_ksi----------------------*/
 delta_ksi=k*(N-1)*0.01;


 /*------------------C[i][j] and B[j] -----------------*/
  for(i=0;i<=N;i++) for(j=0;j<=N;j++) C[i][j]=0;
  for(i=0;i<=N;i++) B[i]=0;
/*---------------------------------------------*/

 INTEGRALS(N-1,p,Iexp,dIexp,I,A,F3,b,c,d);


 ksi2=KSI_2(1.e+08,N-1,Iexp,dIexp,I,A,F3,B,C);
 delta_f3_2=F3_2(N-1,F3,B);

 if( (ksi2<=k*N) && (delta_f3_2<=1.e-04) )    psi=1.e+08;

 else
 { 
printf("ELSE\n"); 
  /*---------- interpol. function( ) ----------*/
  psi=teor_psi(k,N-1,Iexp,dIexp,I,A,F3,B,C);

  /*------------------ delta_psi ------------------------*/
 /*  printf("teor_psi= %e\n",psi); */

   delta_psi=1.0e-10;
   while(psi>delta_psi) delta_psi=10*delta_psi;
   delta_psi=delta_psi/10.;

 /*  printf("delta_psi= %e\n\n",delta_psi); */

  /*------------ psi_left AND psi_right ---------------*/
   psi_right=psi;
   psi_left=psi;

   ksi_right=KSI_2(psi_right,N-1,Iexp,dIexp,I,A,F3,B,C);
   ksi_left=ksi_right;
  /*-------------------------*/
   delta=ksi_right-k*N;
   if(delta<0) delta=-delta;
  /*-------------------------*/

   if(delta>delta_ksi)
   {
    if( ksi_right<k*N )
     while( ksi_right<k*N )
	{ psi_left=psi_right;
	  psi_right=psi_right+delta_psi;

	  ksi_left=ksi_right;
	  ksi_right=KSI_2(psi_right,N-1,Iexp,dIexp,I,A,F3,B,C);
	}
    else
     while( ksi_left>k*N )
	{ psi_right=psi_left;
	  psi_left=psi_left-delta_psi;
	  if(psi_left<0) psi_left=0;

	  ksi_right=ksi_left;
	  ksi_left=KSI_2(psi_left,N-1,Iexp,dIexp,I,A,F3,B,C);
	}
    }

  /*--------------- dikhotomia ------------------------*/

   psi=(psi_left+psi_right)/2.0;
   ksi_psi=KSI_2(psi,N-1,Iexp,dIexp,I,A,F3,B,C);

   delta=ksi_psi-k*N;
   if(delta<0) delta=-delta;


   while(delta>delta_ksi)
   {
     psi=(psi_left+psi_right)/2.0;
     ksi_psi=KSI_2(psi,N-1,Iexp,dIexp,I,A,F3,B,C);
     /*-------------------------*/
     delta=ksi_psi-k*N;
     if(delta<0)
       { delta=-delta;
	 psi_left=psi;
	}

     else  psi_right=psi;

    }
 }
/*------------ psi and ksi-------------------*/

  ksi2=KSI_2(psi,N-1,Iexp,dIexp,I,A,F3,B,C);
  delta_f3_2=F3_2(N-1,F3,B);

/*------------------ I[i] ------------------------*/
  I[0]=B[0];
  I[1]=B[1]+1./8*B[0]+1./8.*B[2];
  I[2]=B[2]+8./37.*B[1]+9./37.*B[3];

  for(i=3;i<=N-3;i++) I[i]=B[i]+1./4.*B[i-1]+1./4.*B[i+1];

  I[N-2]=B[N-2]+9./37.*B[N-3]+8./37.*B[N-1];
  I[N-1]=B[N-1]+1./8.*B[N-2]+1./8.*B[N];
  I[N]=B[N];

/*---------- end -----------------*/

  for(j=0;j<=N;j++) free(A[j]);   free(A);
  for(j=0;j<=N;j++) free(F3[j]);  free(F3);
  free(B);
  for(j=0;j<=N;j++) free(C[j]);   free(C);
  free(b);  free(c);free(d);
}

#ifndef __SPLINE__

void main(int argc,char** argv)
{
  int i;

  double *Iexp,*dIexp,*I,*b,k;

  char procName[60],xName[60],yName[60];
  double xMin,xMax;
  int dim;

  if(argc==1) k=1;
  else sscanf(argv[1],"%lf",&k);


  scanf("%[^\n]\n",procName);

  scanf("x-axis: \"%[^\"]\" from %lf to %lf N_bins= %d\n",xName,&xMin,&xMax,&dim);
  scanf("%[^\n]\n",yName);

  Iexp=(double *)malloc((dim) *sizeof(double));
  dIexp=(double *)malloc((dim) *sizeof(double));

  for(i=0;i<dim ;i++)
    if(2!=scanf("%lf +/- %lf",Iexp+i,dIexp+i)) 
    fprintf(stderr,"Wrong data format\n");


  I=(double *)malloc((dim+1) *sizeof(double));

  b=(double *)malloc((dim+1) *sizeof(double));

/* -------------------SPLINE----------------------------------------------*/

  spline3smooth(k,1.,dim,Iexp,dIexp,I);
  spline3make(dim+1,I,b);
 
 /*------------ writing to the file -------------------------------------*/

   printf("%s\n",procName);
   printf("x-axis: \"%s\" from %e to %e N_points= 201\n",xName,xMin,xMax);
   printf("%s\n",yName);

   for(i=0;i<201;i++) printf("%e\n",  spline3val(dim+1,I,b, i*dim/200.,NULL) );

   free((void*)Iexp);
   free((void*)dIexp);
   free((void*)I);
   free((void*)b);
}
#endif
