#include <stdlib.h>
#include <stdio.h>
#include <math.h>

#define INCUB(x)   ((x)>0   ? (0.5+(x))/((x)+1.) : 0.5/(1.-(x)))
#define OUTCUB(x)  ((x)>0.5 ? (0.5-(x))/((x)-1.) : 1-0.5/x)

            
/*======================================================*/
static double amotry(double **p, double *y, int ndim,
	     double (*f)(double *), int ilo, double fac)
{
   int i,j;
   double fac1=(1.0-fac)/ndim;
   double ytry;

   for(j=0;j<ndim;j++) p[ndim+1][j]=p[ilo][j]*fac;
   for(i=0;i<=ndim;i++) 
   if(i!=ilo) for(j=0;j<ndim;j++) p[ndim+1][j]+=p[i][j]*fac1; 

   ytry=(*f)(p[ndim+1]);
   if (ytry > y[ilo]) 
   {  double * pswap=p[ndim+1];
      p[ndim+1]=p[ilo];
      p[ilo]=pswap;
      y[ilo]=ytry;
   }
   return ytry;
}


static void amoeba(double **p, double * y, int ndim, double (*f)(double *), 
                                                    double eps, int *nCalls)
{
   int i,ilo,ihi,inlo,j;
   double ysave,ytry;

   for (;;) 
   {
      ihi=0;									     
     										     
      ilo = y[0]<y[1] ? (inlo=1,0) : (inlo=0,1);				     
      for (i=0;i<=ndim;i++)							     
      {										     
     	 if (y[i] >= y[ihi]) ihi=i;
     	 if (y[i] < y[ilo]) { inlo=ilo; ilo=i; } 
     	 else if (y[i] < y[inlo] && i != ilo) inlo=i;
      }										     
     										     
      if((*nCalls)<=0||2*(y[ihi]-y[ilo])/(fabs(y[ilo])+fabs(y[ihi]))<eps)break;
     										     
      ytry=amotry(p,y,ndim,f,ilo,-1.0); (*nCalls)--;				     
      if (ytry >= y[ihi]) {ytry=amotry(p,y,ndim,f,ilo,2.); (*nCalls)--;}	     
      else if (ytry <= y[inlo])							     
      {										     
         ysave=y[ilo];								     
     	 ytry=amotry(p,y,ndim,f,ilo,0.5);  (*nCalls)--;
     	 if (ytry <= ysave)
     	 {  
printf(" shrink\n");     	 
     	    for (i=0;i<=ndim;i++)
     	    { 
               if (i != ihi)
     	       {
     		  for(j=0;j<ndim;j++) p[i][j]=0.5*(p[i][j]+p[ihi][j]);
     		  y[i]=(*f)(p[i]);  		  
     	       }
            }	            								     
     	    (*nCalls) -= ndim;
         }									     
      }										     
   }
   { double swap=y[0];							     
     double *pswap=p[0];							     
     y[0]=y[ihi]; y[ihi]=swap;						     
     p[0]=p[ihi]; p[ihi]=pswap;						     
   }										     
}
                    

/*========================== end of amoeba ================*/

static double ff(double * x)
{  double xx[2];
   double y,x2; 
   
   xx[0]=INCUB(x[0]);
   xx[1]=INCUB(x[1]);
   x2=(xx[0]+xx[1]-0.5);
   y= 1/(xx[0]+1)/(x2*x2+0.001);
   printf("x=(%f,%f) y=%f\n",xx[0],xx[1],y);
   return y;
}

int main(void)
{
  double *p[4];
  double y[4];
  int i,j;
  
  int nCall=100;
  for(i=0;i<4;i++)p[i]=malloc(2*sizeof(double));
  
  p[0][0]=10;p[0][1]=10;
  p[1][0]=10;p[1][1]=11;
  p[2][0]=11;p[2][1]=10;


  for(i=0;i<=2;i++)  y[i]=ff(p[i]);
    
  amoeba(p,y,2,ff,0.001,&nCall);
}
